import copy
import random
import sys

from microgp4 import microgp4
from microgp4.ctg.constraints.constraints import Constraints
from microgp4.ctg.parameters import reference_type
from microgp4.ctg.section import sub_section_key, section_key
from microgp4.ctg.sectioned_tagged_graph import SectionedTaggedGraph, node_key
from microgp4.exception import Microgp4ViolationConstraint

sections_occurrence_key = "sections_occurrence"
node_number_key = "node_number"
first_node_key = "first_node"
last_node_key = "last_node"


class ConstrainedTaggedGraph(SectionedTaggedGraph):
    """
    This is meant to be the solution or individual class.
    """

    def __init__(self, constraints_name, meta_section, data=None, **attr):
        """
        Parameters
        ----------
        constraints_name: str
                        The name of the constraints which this solutions will use
        meta_section: Section
                    The section contain all the information and constraints to build a ConstrainedTaggedGraph
        """
        self.meta_section = meta_section
        self._sections_occurrence = dict()
        self.__constraints = Constraints(constraints_name)
        self.__node_to_finalize = list()
        SectionedTaggedGraph.__init__(self, data, **attr)

    def _update_section_global_occurrence(self, section_type):
        """
        Given the section type the function will increment the internal count
        and return the number of the current section of the given type.
        This function is used for internal purpose, and it should never be called.

        :param section_type: The type of the section, the name in the JSON file and contained in the meta-section.
        :type section_type: str
        :return: Current number of section of the given type.
        """
        occurrence = self._sections_occurrence.pop(section_type, 0) + 1
        self._sections_occurrence[section_type] = occurrence
        return occurrence

    def section_global_occurrence(self, section_type):
        """
        The method will return the total number of section, in the current solution, of the given section type

        :param section_type: The type of a section
        :type section_type: str
        :return: Section type occurrence in the ctg
        """
        if section_type not in self._sections_occurrence:
            return 0
        else:
            return self._sections_occurrence[section_type]

    def _update_section_occurrence(self, section, section_type):
        """
        Given the section type the function will increment the internal count inside the section
        and return the number of the current section of the given type.
        This function is used for internal purpose, and it should never be called.

        :param section: The section ID.
        :param section_type: The section type.
        :return: Section type occurrence in the given section.
        """
        occurrence_dict = self.section[section][sections_occurrence_key]
        occurrence = occurrence_dict.pop(section_type, 0) + 1
        occurrence_dict[section_type] = occurrence
        return occurrence

    def section_occurrence(self, section, section_type):
        """
        The method will return the occurrence of a particular section type inside a section

        :param section: The section in which sub sections will be counted
        :type section: Section
        :param section_type: The section type to be counted
        :type section_type: str
        :return: The number of occurrence of the given section type in the section
        """
        s = self.section[section]
        if sections_occurrence_key not in s:
            return 0
        elif section_type not in s[sections_occurrence_key]:
            return 0
        else:
            return s[sections_occurrence_key][section_type]

    def section_type(self, section):
        """
        Given the section the method will return the section type.

        :param section: Section
        :return: Section type
        """
        if isinstance(section, dict):
            return section["type"]
        return self.section[section]["type"]

    def _update_nodes_in_section_number(self, section):
        """
        It Increment the number node count inside the section.
         This function is used for internal purpose, and it should never be called.

        :param section: Section ID.
        :type section: str
        :return: The updated node count.
        """
        s = self.section[section]

        nn = s.pop(node_number_key, 0) + 1
        s[node_number_key] = nn
        return nn

    def nodes_in_section_number(self, section):
        """
        The method returns the number of nodes contained inside the section.
        If the section cannot have any nodes return 0.

        :param section: Section ID.
        :type section: str
        :return: Number of node inside the section.
        """

        s = self.section[section]
        if node_number_key not in s:
            return 0
        else:
            return s[node_number_key]

    def add_section(self, section_type, tags_dict=None, **attr):
        """
        Add new section inside the solution. The ID will be created internally and it is not possible to change.
        Only in debug mode is checked if there is not too much section of the given type.

        :param section_type: The type of the section.
        :type section_type: str
        :param tags_dict:
        :param attr:
        :return: The section ID.
        """
        assert section_type in self.meta_section.section
        section_global_occurrence = self._update_section_global_occurrence(section_type)
        assert self.meta_section.section[section_type]['max'] >= section_global_occurrence

        section_attr = {"type": section_type, sections_occurrence_key: dict()}
        extra_info = dict()
        extra_info.update(attr)
        if tags_dict is not None:
            extra_info.update(tags_dict)

        section_attr["extra_info"] = extra_info
        return SectionedTaggedGraph.add_section(self, section_attr)

    def get_all_section_by_type(self, section_type):
        """
        Return all the sections that have the given type in a list.
        If there are not any section of this type the list is empty.

        :param section_type: The section type.
        :type section_type: str
        :return A list contained all the Section of the given type.
        """
        result = list()

        for _, s in self.section.items():
            if 'type' in s and s['type'] == section_type:
                result.append(s)

        return result

    def remove_section(self, section, remove_all=True):
        """
        Remove section from the current solution. The nodes inside the section will not be deleted,
        but they become nodes without any section.

        :param section: Section ID.
        :type section: str
        :param remove_all: If True all the sub sections (if any) are removed as well.
        :type remove_all: bool
        """
        section_type = self.section_type(section)
        self._sections_occurrence[section_type] = max(self._sections_occurrence[section_type] - 1, 0)
        SectionedTaggedGraph.remove_section(self, section, remove_all)

    def put_section_inside_section(self, section, inner_section):
        """
        The method puts an existing section inside and other existing section, the two section must be different.

        :param section: The id of the section which will contain the other section.
        :param inner_section: The id of the section will be put inside the outer one.
        """

        assert node_key not in self.section[section]

        SectionedTaggedGraph.put_section_inside_section(self, section, inner_section)
        can_have_sub_section = False
        for s in self.meta_section.section[self.section_type(section)][sub_section_key]:
            if s["name"] == self.section_type(inner_section):
                section_occurrence = self._update_section_occurrence(section, self.section_type(inner_section))
                assert s["max"] >= section_occurrence
                can_have_sub_section = True
                break
        assert can_have_sub_section

    def remove_section_from_section(self, section, inner_section):
        """
        It remove a section from the outer section, but the section will no be deleted.

        :param section: The id of the section which contain the one to remove.
        :param inner_section:  The id of the section that will be removed from the other section.
        """
        SectionedTaggedGraph.remove_section_from_section(self, section, inner_section)
        occurrence_dict = self.section[section][sections_occurrence_key]
        section_type = self.section_type(inner_section)
        occurrence_dict[section_type] = max(occurrence_dict[section_type] - 1, 0)

    def create_new_section(self, section_type):
        """
        Create a new section similar to "add_section" but also put it inside a section if needed and create
        upper sections if not present. The section created is empty.

        :param section_type: The section type
        :type section_type: str
        :return: The created section ID
        """
        assert section_type in self.meta_section.section

        meta_section = self.meta_section.section[section_type]
        if "section" in meta_section:
            outer_sections_id = meta_section["section"]
            outer_section_id = None

            if outer_sections_id is not None and len(outer_sections_id) > 0:
                outer_section_type = random.choice(outer_sections_id)
                outer_sections = self.get_all_section_by_type(outer_section_type)

                if outer_sections is None or len(outer_sections) is 0:
                    outer_sections = [self.section[self.create_new_section(outer_section_type)]]

                outer_section_value = random.choice(outer_sections)

                for s_key, s_value in self.section.items():
                    if outer_section_value is s_value:
                        outer_section_id = s_key
                        break
                assert outer_section_id is not None

            created_section = self.add_section(section_type)

            if outer_section_id is not None:
                self.put_section_inside_section(outer_section_id, created_section)

            return created_section

    def add_node(self, section=None, macro_group=None, macro_id=None, tags_dict=None, **attr):
        """
        Add a new node in the solution and generate all the object for the parameters.
        Each of the arguments are optional, if not specified they will be automatically generated.
        In case of bad random generation the method will return False.
        An assert will occur if the inputs are given wrong.

        :type section: str
        :type macro_group: str
        :type macro_id: int
        :type tags_dict: dict
        :type attr: dict
        :param section: the section name.
        :param macro_group: the group name.
        :param macro_id: macro identifier.
        :param tags_dict: extra node info.
        :param attr: extra node info.
        :return: the node name, or False if the method fails.
        """

        if section is None:
            section = random.choice(list(self.section))
        else:
            assert section in self.section

        s = self.section[section]

        if sub_section_key in s:
            return False

        s_type = self.section_type(section)
        nodes = self.meta_section.section[s_type]["nodes"]

        if s[node_number_key] >= nodes["max"]:
            return False

        macros = nodes["macros"]

        if macro_group is None:
            macro_group = random.choice(list(macros))
        else:
            assert macro_group in macros

        m_g = macros[macro_group]

        if macro_id is None:
            macro_id = random.choice(list(m_g))
        else:
            assert macro_id in m_g

        if node_key not in s or first_node_key not in s:
            node = self.append_node(section, macro_group, macro_id, tags_dict=tags_dict, **attr)
            try:
                self.finalize_nodes()
                return node
            except:
                self.remove_node(node)
                return False

        node_list_in_section = s[node_key]

        last_position = type("last_node", (), {})()

        index = random.choice(node_list_in_section + [last_position, ])

        if index is last_position:
            node = self.append_node(section, macro_group, macro_id, tags_dict=tags_dict, **attr)
            try:
                self.finalize_nodes()
                return node
            except:
                self.remove_node(node)
                return False

        self._update_nodes_in_section_number(section)

        if tags_dict is not None:
            node_tag = tags_dict
        else:
            node_tag = dict()

        node = SectionedTaggedGraph.add_node(self, section, **attr)

        if index == s[first_node_key]:

            old_first = s[first_node_key]
            s[first_node_key] = node
            if self.fan_in_node(old_first, False, True) > 1:
                previous_node = self.__get_previous_node(old_first)
                for node_id in previous_node:
                    node_val = self.nodes[node_id]
                    parameters = node_val['macro']['parameters']

                    for parameter_key, parameter_value in parameters.items():
                        if issubclass(parameter_value, reference_type.ReferenceType) and parameter_value.is_outer():
                            parameter_value.value = node

            self.add_edge(node, old_first, type="next")
        else:
            successor = None
            predecessor = None

            to_break = False

            p_list = self.predecessors(index)
            for p in p_list:
                for e, v in self.adj[p][index].items():
                    if "type" in v:
                        if v["type"] == "next":
                            predecessor = p
                            successor = index
                            self.remove_edge(predecessor, successor, e)
                            to_break = True
                            break

                if to_break:
                    break

            assert successor is not None and predecessor is not None

            self.add_edge(predecessor, node, type="next")
            self.add_edge(node, successor, type="next")

        node_tag.update({"macro": {"group": macro_group, "id": macro_id, "parameters": dict()}})
        try:
            for k, c in nodes["macros"][macro_group][macro_id]["parameters"].items():
                node_tag["macro"]["parameters"][k] = c(node)
        except:
            self.remove_node(node)
            return False

        self.add_tags_to_node(node, node_tag)

        return node

    def append_node(self, section, macro_group, macro_id, tags_dict=None, **attr):
        """
        Similar to 'add_node', this method will create and add a new node in the last position in the given section.
        Some generated nodes might need to bi finalized.

        :type section: str
        :type macro_group: str
        :type macro_id: int
        :type tags_dict: dict
        :type attr: dict
        :param section: the section name.
        :param macro_group: the group name.
        :param macro_id: macro identifier.
        :param tags_dict: extra node info.
        :param attr: extra node info.
        :return: the node name, or False if the method fails.
        """
        s = self.section[section]
        assert sub_section_key not in s
        s_type = self.section_type(section)

        nodes = self.meta_section.section[s_type]["nodes"]
        nodes_in_section_number = self._update_nodes_in_section_number(section)
        assert nodes["max"] >= nodes_in_section_number

        node_tag = dict()

        if tags_dict is not None:
            extra_tag = tags_dict
        else:
            extra_tag = dict()

        extra_tag.update(attr)

        node = SectionedTaggedGraph.add_node(self, section, extra_tag=extra_tag)

        node_tag.update({"macro": {"group": macro_group, "id": macro_id, "parameters": dict()}})
        for k, c in nodes["macros"][macro_group][macro_id]["parameters"].items():

            if issubclass(c, reference_type.ReferenceType):
                self.__node_to_finalize.append(
                    {'node_parameters': node_tag["macro"]["parameters"], 'class': c, 'par_name': k, 'node_id': node})
            else:
                node_tag["macro"]["parameters"][k] = c(node)

        self.add_tags_to_node(node, node_tag)

        if first_node_key not in s:
            s[first_node_key] = node
            s[last_node_key] = node

        if s[last_node_key] is not node:
            self.add_edge(s[last_node_key], node, type="next")
            s[last_node_key] = node

        return node

    def finalize_nodes(self):
        """
        This method will finalize all node. To be called after one or more 'append_node'.

        """
        for n in self.__node_to_finalize:
            n['node_parameters'][n['par_name']] = n['class'](n['node_id'])
        self.__node_to_finalize = list()

    def fan_in_section(self, section_id):
        """
        Return the number of in-put link the given section has.

        :param section_id: The section ID.
        :type section_id: str
        :return: Fan in of the given Section.
        """

        section_dict = self.section[section_id]

        if sub_section_key in section_dict:
            return 0

        section_first_node = section_dict["first_node"]
        predecessor_list = self.__get_previous_node(section_first_node)

        predecessor_count = 0

        for node in predecessor_list:
            edge = self.adj[node][section_first_node]
            for edge_id, edge_value in edge.items():
                if edge_value["outer"] == "true":
                    predecessor_count += 1

        return predecessor_count

    def fan_in_node(self, node_id, count_next_adj=False, count_section_ref=False):
        """
        It returns the number of in-put edge for the given node.
        The some edge types can be excluded from the count.

        :param node_id: The node ID.
        :type node_id: str
        :param count_next_adj: If True the next arch will be counted
        :type count_next_adj: bool
        :param count_section_ref: If True also the edges referring to the section are counted.
        :return: The number of in edge, according to the constraints.
        """
        predecessor_list = self.__get_previous_node(node_id)

        predecessor_count = len(predecessor_list)

        if predecessor_count is 0:
            return 0
        else:
            section_first_node = self.section[self.nodes[node_id][section_key]]["first_node"]
            need_to_count_all = not count_section_ref or section_first_node != node_id
            if count_next_adj and need_to_count_all:
                return predecessor_count
            else:
                found_next = False
                for e in predecessor_list:
                    edge = self.adj[e][node_id]
                    for edge_id, edge_value in edge.items():
                        if edge_value['type'] == "next":
                            found_next = not need_to_count_all
                            predecessor_count -= 1
                            if found_next:
                                break
                        elif edge_value["outer"] == "true":
                            predecessor_count -= 1
                    if found_next:
                        break

            return predecessor_count

    def fan_out(self, node_id, count_next_adj=False):
        """
        It returns the fan-out for a node.
        The type "next" can be excluded from the count.

        :param node_id: Node ID.
        :type node_id: str
        :param count_next_adj: If True the next arch will be counted.
        :type count_next_adj: bool
        :return: The number of outgoing edge, according to the constraints.
        """
        successor_list = self.__get_successors_node(node_id)
        successor_count = len(successor_list)

        if successor_count is 0:
            return 0

        if count_next_adj:
            return successor_count
        else:
            found_next = False
            for e in successor_list:
                edge = self.adj[node_id][e]
                for edge_id, edge_value in edge.items():
                    if edge_value['type'] == "next":
                        found_next = True
                        successor_count -= 1
                        break
                if found_next:
                    break
            return successor_count

    def remove_node(self, node):
        """
        It will completely delete the given node and all its in/out arch.

        :param node: The node ID.
        :type node: str
        """
        section = self.section[self.nodes[node][section_key]]
        section_id = self.nodes[node][section_key]

        if node not in (section[first_node_key], section[last_node_key]):
            p_list = self.predecessors(node)
            s_list = self.successors(node)

            predecessor = None
            successor = None

            for p in p_list:
                to_end = False
                for _, v in self.adj[p][node].items():
                    if "type" in v:
                        type_ = v["type"]
                        if type_ is "next":
                            predecessor = p
                            to_end = True
                            break
                if to_end:
                    break

            for s in s_list:
                to_end = False
                for _, v in self.adj[node][s].items():
                    if "type" in v:
                        type_ = v["type"]
                        if type_ is "next":
                            successor = s
                            to_end = True
                            break
                if to_end:
                    break
            if predecessor is not None and successor is not None:
                self.add_edge(predecessor, successor, type="next")
            elif predecessor not in (section[first_node_key], section[last_node_key]) or successor not in \
                    (section[first_node_key], section[last_node_key]):
                raise Microgp4ViolationConstraint("There is no edge 'next' that connect the node '{0}' "
                                                  "with other node inside section "
                                                  "'{1}'!".format(node, self.nodes[node][section_key]))
            SectionedTaggedGraph.remove_node(self, node)
        elif node == section[first_node_key] and node == section[last_node_key]:
            SectionedTaggedGraph.remove_node(self, node)
            self.remove_section(section_id)
        elif node == section[first_node_key] or node == section[last_node_key]:
            if node == section[first_node_key]:
                successor = None
                s_list = self.successors(node)
                for s in s_list:
                    to_end = False
                    for _, v in self.adj[node][s].items():
                        if "type" in v:
                            type_ = v["type"]
                            if type_ is "next":
                                successor = s
                                to_end = True
                                break
                    if to_end:
                        break
                if successor is None:
                    raise Microgp4ViolationConstraint(
                        "Node '{0} has not successor'".format(node, self.nodes[node][section_key]))
                section[first_node_key] = successor

            if node == section[last_node_key]:
                predecessor = None
                p_list = self.predecessors(node)
                for p in p_list:
                    to_end = False
                    for _, v in self.adj[p][node].items():
                        if "type" in v:
                            type_ = v["type"]
                            if type_ is "next":
                                predecessor = p
                                to_end = True
                                break
                    if to_end:
                        break
                if predecessor is None:
                    raise Microgp4ViolationConstraint(
                        "Node '{0} has not predecessor'".format(node, self.nodes[node][section_key]))
                section[last_node_key] = predecessor

            SectionedTaggedGraph.remove_node(self, node)

    def __get_previous_node(self, node_id):
        """
        Exist only because in same case a list is more useful instead of an Iterator.

        :param node_id: The Node ID.
        :return: The List of the predecessor of the given node.
        """
        return list(self._pred[node_id])

    def __get_successors_node(self, node_id):
        """
        Exist only because in same case a list is more useful instead of an Iterator.

        :param node_id: The Node ID.
        :return: The List of the successors of the given node.
        """
        return list(self._succ[node_id])

    def __get_node(self, result_list, node_id, node_function):
        """
        Internal function used both to get the previous and the successors nodes, based on the given node function.

        :param result_list: The list to be filled with the appropriate nodes.
        :type result_list: list
        :param node_id: The initial node ID.
        :type node_id: str
        :param node_function: The function which will retrieve the wanted nodes.
        :return: The list contained the wanted nodes
        """

        section_node_list = self.section[self.nodes[node_id][section_key]][node_key]

        to_return = False

        while True:
            if node_id not in section_node_list:
                return

            node_to_check = node_function(self, node_id)

            if node_to_check is None or len(node_to_check) is 0:
                return

            if to_return:
                return

            to_return = True

            for n in node_to_check:

                to_break = False

                if node_function is ConstrainedTaggedGraph.__get_previous_node:
                    edge_dic = self.adj[n][node_id]
                else:
                    edge_dic = self.adj[node_id][n]

                for k, v in edge_dic.items():
                    if 'type' in v:
                        if v['type'] == 'next':
                            to_return = False
                            result_list.append(n)
                            node_id = n
                            to_break = True
                            break

                if to_break:
                    break

    def get_previous_node(self, node_id):
        """
        Get all the previous nodes of the given one, inside the same section and all connected with next edge.

        :param node_id: The node ID.
        :type node_id: str
        :return: The list of the previous nodes ids. It will contain always the given node.
        """
        result = [node_id, ]
        self.__get_node(result, node_id, ConstrainedTaggedGraph.__get_previous_node)
        return result

    def get_successors_node(self, node_id):
        """
        Get all the previous nodes of the given one, inside the same section and all connected with next edge.

        :param node_id: The node ID.
        :type node_id: str
        :return: The list of the successors nodes ids. It will contain always the given node.
        """
        result = [node_id, ]
        self.__get_node(result, node_id, ConstrainedTaggedGraph.__get_successors_node)
        return result

    def __check_reference_format(self):
        """
        Internal function to check the reference format.

        :return: The correctness of reference format.
        :rtype: bool
        """

        def check_reference(adj):
            for source in adj:
                for dest in adj[source]:
                    for edge_id in adj[source][dest]:
                        edge = adj[source][dest][edge_id]
                        if edge["type"] == "ref":
                            if "outer" not in edge or edge["outer"] not in ("true", "false"):
                                return False
            return True

        return check_reference(self._succ) and check_reference(self._pred)

    def check(self):
        """
        This method will check if the solution is feasible, using the constraints.

        :return: The check result.
        :rtype:bool
        """

        assert self.__check_reference_format()

        try:
            self.__constraints.exec_constraints(self)
            return True
        except Microgp4ViolationConstraint as err:
            self.log.exception("{0}".format(err))
            return False

    def _dump_node(self, section):
        """
        Internal function use to convert avery nodes of a section in text.

        :param section: The section.
        :type section: dict
        :return: A list with the all the nodes converted to string.
        """

        label_set = set()

        if node_key not in section:
            return
        else:
            section_type = self.section_type(section)
            node_id = section[first_node_key]
            last_node_id = section[last_node_key]
            printed_text_list = list()

            while True:

                if node_id not in section[node_key]:
                    return printed_text_list

                node = self.nodes[node_id]

                node_macro = node['macro']
                group = node_macro['group']
                macro_id = node_macro["id"]
                expression = self.meta_section.section[section_type]['nodes']['macros'][group][macro_id]["expression"]

                par = node_macro['parameters']

                predecessors_list = self.predecessors(node_id)

                for predecessor in predecessors_list:
                    edge = self.adj[predecessor][node_id]
                    for _, edge_type in edge.items():
                        if edge_type["type"] == "next":

                            predecessor_node = self.nodes[predecessor]

                            predecessor_par = predecessor_node['macro']['parameters']

                            for k, p in predecessor_par.items():
                                if isinstance(p, reference_type.ReferenceType):
                                    if not p.is_outer():
                                        if p.get_label() not in label_set:
                                            label_set.add(p.get_label())
                                            printed_text_list.append(p.get_label())

                text = expression.format(**par)
                printed_text_list.append(text)

                if node_id == last_node_id:
                    return printed_text_list
                successor_list = self.successors(node_id)

                for successor_node in successor_list:
                    to_break = False
                    tmp_edge = self.adj[node_id][successor_node]
                    for k, v in tmp_edge.items():
                        if 'type' in v:
                            if v['type'] == 'next':
                                node_id = successor_node
                                to_break = True
                                break

                    if to_break:
                        break

    def _visit(self, starting_section, depth_first=True):
        """
        Visit the graph.

        :param starting_section: The staring section of the visit.
        :type starting_section: dict
        :param depth_first: It determines the visit type.
        :return: A list of nodes converted in text.
        """
        queue = [starting_section]

        printed_tex_list = list()

        while len(queue) > 0:
            section = queue.pop(0)
            if sub_section_key not in section:
                printed_tex_list += self._dump_node(section)
            else:
                section_list_name = section[sub_section_key]

                sub_section_list = [self.section[s] for s in section_list_name]

                if depth_first:
                    queue = sub_section_list + queue
                else:
                    # breadth first
                    queue = queue + sub_section_list
        return printed_tex_list

    def dump_solution(self, file_dump=sys.stdout, path=None, depth_first=True):
        """
        This method will dump the current solution, converting it into text.
        It accept or a pointer to an open file, or a path in which create or override a file.

        :param file_dump: The file object.
        :type file_dump: object
        :param path: The file path. The file can be existing or not.
        :type path: str
        :param depth_first: The type of visit of the graph.
        :type depth_first: bool
        """

        assert path is None or file_dump is sys.stdout

        to_close = False
        root = None

        if path is not None:
            file_dump = open(path, "w+")
            to_close = True

        for _, s in self.section.items():
            if self.section_type(s) == "root":
                root = s
                break

        assert root is not None
        solution_text_list = self._visit(root, depth_first)

        print(*solution_text_list, sep='\n', file=file_dump)

        if to_close is not None:
            file_dump.close()

    def clone_solution(self):
        """
        This method will clone the entire solution, content itself can have some difference from the original one,
        for example the IDs will change, but the result of the dump is identical, also the number of section, nodes,
        type of section, nodes and their order are the same.


        :return: The cloned solution.
        :rtype: ConstrainedTaggedGraph
        """
        sections_mapping = dict()
        nodes_mapping = dict()

        ctg_cloned = ConstrainedTaggedGraph(self.__constraints.constraints_name, self.meta_section,
                                            **copy.deepcopy(self.graph))

        microgp4.add_new_solution(ctg_cloned)

        try:
            for section_id, section in self.section.items():
                sections_mapping[section_id] = ctg_cloned.add_section(section["type"],
                                                                      copy.deepcopy(section["extra_info"]))

            for section_id, section in self.section.items():
                section_cloned_id = sections_mapping[section_id]
                if "sub_sections" in section:
                    for sub_section in section["sub_sections"]:
                        ctg_cloned.put_section_inside_section(section_cloned_id, sections_mapping[sub_section])
                elif "node" in section:
                    for node_id in section["node"]:
                        original_node = self.nodes[node_id]
                        macro = original_node["macro"]
                        new_node_id = ctg_cloned.append_node(section_cloned_id, macro["group"], macro["id"],
                                                             copy.deepcopy(original_node["extra_tag"]))

                        new_parameters = ctg_cloned.nodes[new_node_id]["macro"]["parameters"]

                        for parameter_name, parameter in macro["parameters"].items():
                            if not isinstance(parameter, reference_type.ReferenceType):
                                new_parameters[parameter_name].value = parameter.value

                        nodes_mapping[node_id] = new_node_id
                        nodes_mapping[new_node_id] = node_id

            for n in ctg_cloned.__node_to_finalize:
                original_node_ref_id = \
                    self.nodes[nodes_mapping[n['node_id']]]['macro']['parameters'][n['par_name']].value[
                        0]
                n['node_parameters'][n['par_name']] = n['class'](n['node_id'], nodes_mapping[original_node_ref_id])
            ctg_cloned.__node_to_finalize = list()
        except:
            microgp4.remove_solution(ctg_cloned)
            raise

        return ctg_cloned
