import logging

from microgp4.ctg.constraints.constraints_type import FanInType
from microgp4.ctg.section import sub_section_key, section_key
from microgp4.ctg.sectioned_tagged_graph import node_key
from microgp4.exception import Microgp4ViolationConstraint

__log = logging.getLogger("microgp4")


def check_occurrences(ctg):
    """
    Check if there are too many sections and nodes or too few.
    In case some constraints are violated, a Microgp4ViolationConstraint will be raise.

    :param ctg: The solution on which the test will be performed.
    :type ctg: ConstrainedTaggedGraph

    """
    meta_section = ctg.meta_section
    for s, val in ctg.section.items():
        g_s_occ = ctg.section_global_occurrence(ctg.section_type(s))
        g_min_occ = meta_section.section[ctg.section_type(s)]["min"]
        g_max_occ = meta_section.section[ctg.section_type(s)]["max"]
        if g_s_occ < g_min_occ:
            error_string = "Section '{0}' does not appear enough times in the solution! " \
                           "times: '{1}' min times {2}".format(s, g_s_occ, g_min_occ)
            __log.error(error_string)
            raise Microgp4ViolationConstraint(error_string)

        if g_s_occ > g_max_occ:
            error_string = "Section '{0}' appears too many times in the solution! " \
                           "times: '{1}' max times '{2}'".format(s, g_s_occ, g_max_occ)
            __log.error(error_string)
            raise Microgp4ViolationConstraint(error_string)

        if sub_section_key in val:
            for e in val[sub_section_key]:
                s_occ = ctg.section_occurrence(s, ctg.section_type(e))
                min_occ = -1
                max_occ = -1
                for el in meta_section.section[ctg.section_type(s)][sub_section_key]:
                    if el["name"] == ctg.section[e]['type']:
                        min_occ = el["min"]
                        max_occ = el["max"]
                        break

                if min_occ < 0:
                    error_string = "In section '{0}' there is no sub section called '{1}'!".format(s, e)
                    __log.error(error_string)
                    raise Microgp4ViolationConstraint(error_string)

                if s_occ < min_occ:
                    error_string = "In section '{0}' there is not enough section '{1}'!" \
                                   "times: '{2}' min times: '{3}'".format(s, e, s_occ, min_occ)

                    __log.error(error_string)
                    raise Microgp4ViolationConstraint(error_string)

                if s_occ > max_occ:
                    error_string = "In section '{0}' there are too many section '{1}' !" \
                                   "times: '{2}' max times: '{3}'".format(s, e, s, max_occ)
                    __log.error(error_string)
                    raise Microgp4ViolationConstraint(error_string)

        elif node_key in val:
            n_min = meta_section.section[ctg.section_type(s)]["nodes"]["min"]
            n_max = meta_section.section[ctg.section_type(s)]["nodes"]["max"]
            n_n = ctg.nodes_in_section_number(s)

            if n_n < n_min:
                error_string = "Section '{0}' does not have enough nodes! " \
                               "nodes number: '{1}' min number '{2}'".format(s, n_n, n_min)
                __log.error(error_string)
                raise Microgp4ViolationConstraint(error_string)

            if n_n > n_max:
                error_string = "Section '{0}' has too many nodes! " \
                               "nodes number: '{1}' man number: '{2}'".format(s, n_n, n_max)
                __log.error(error_string)
                raise Microgp4ViolationConstraint(error_string)


def check_parameters(ctg):
    """
    Check if all nodes contain the correct parameters.
    In case some constraints are violated, a Microgp4ViolationConstraint will be raise.

    :param ctg: The solution on which the test will be performed.
    :type ctg: ConstrainedTaggedGraph

    """
    meta_section = ctg.meta_section
    nodes = ctg.node

    for n_id, node in nodes.items():
        node_section = node[section_key]
        node_macro = node["macro"]
        group = node_macro['group']
        macro_id = node_macro['id']
        macro_parameters = meta_section.section[ctg.section_type(node_section)]['nodes']['macros'][group][macro_id][
            "parameters"]

        for k, v in node_macro['parameters'].items():
            if not isinstance(v, macro_parameters[k]):
                error_string = "Node id '{0}' has parameters type: '{1}', " \
                               "instead of: '{2}'!".format(n_id, type(v), type(macro_parameters[k]))

                __log.error(error_string)
                raise Microgp4ViolationConstraint(error_string)

            if not v.check():
                error_string = "Parameter '{0}' of node id '{1}' " \
                               "has wrong value: '{2}'!".format(k, n_id, v.value)

                __log.error(error_string)
                raise Microgp4ViolationConstraint(error_string)


def check_fan_in(ctg):
    """
    Check if all parameters have the correct fan-in.
    In case some constraints are violated, a Microgp4ViolationConstraint will be raise.

    :param ctg: The solution on which the test will be performed.
    :type ctg: ConstrainedTaggedGraph

    """
    meta_section = ctg.meta_section.section
    nodes = ctg.node
    sections = ctg.section

    for section_id, section_value in sections.items():
        section_type = section_value["type"]
        section_fan_in = ctg.fan_in_section(section_id)
        expected_fan_in = meta_section[section_type]["fan_in"].value
        if expected_fan_in != FanInType.INFINITE.value and section_fan_in > expected_fan_in:
            error_string = "Section '{}' has fan in '{}' instead of '{}'.".format(section_id, section_fan_in,
                                                                                  expected_fan_in)
            __log.error(error_string)
            raise Microgp4ViolationConstraint(error_string)

    for node_id in nodes:
        node = ctg.node[node_id]
        node_macro = node["macro"]
        section_macros = meta_section[sections[node["section"]]["type"]]["nodes"]["macros"]
        expected_node_fan_in = section_macros[node_macro["group"]][node_macro["id"]]["fan_in"]
        node_fan_in = ctg.fan_in_node(node_id)

        if expected_node_fan_in is not FanInType.INFINITE and node_fan_in > expected_node_fan_in.value:
            error_string = "Node '{}' has fan in '{}' instead of '{}'.".format(node_id, node_fan_in,
                                                                               expected_node_fan_in.value)
            __log.error(error_string)
            raise Microgp4ViolationConstraint(error_string)


_default_function = {check_occurrences, check_parameters, check_fan_in}


class Constraints:
    """
    A wrapper class of the Pool Constraints class.
    In this way in the program exist only a object containing all the pools.

    __constrains is the static private global object containing all the pools.
    """
    __constrains = None

    class __Constraints:
        """
        The actual class containing all the constraints pool.
        It provides to a solution a specific constraints poll or it create a new one,
        if the request pool does not exist.
        """

        def __init__(self):
            self.constraints_dict = dict()

        def get_constraints(self, constraints_name):
            """
            Giving a name a constraints pool is returned. If there is not a pool with the provided name,
            a new pool is created.

            :param constraints_name: Unique pool name
            :return: the wanted pool, if it does not exist a new one is created
            """
            if constraints_name not in self.constraints_dict:
                self.constraints_dict[constraints_name] = set()
                self.constraints_dict[constraints_name].update(_default_function)
            return self.constraints_dict[constraints_name]

        def insert_constraint(self, constraints_name, constraint_function):
            """
            Inject a new constraint function in a certain pool, that must exist.

            All the functions must have only one input parameters (the solution),
            and they do not return anything, instead they raise the appropriate exception.

            :param constraints_name: The unique name of the pool
            :param constraint_function: The new constraints function
            :return: The number of function inside the pool
            """
            self.constraints_dict[constraints_name].add(constraint_function)
            return len(self.constraints_dict[constraints_name])

        def exec_constraints(self, constraints_name, solution_ctg):
            """
            It executes all the constraint in the wanted pool on a certain solution.
            The pool must exist.

            :param constraints_name: The unique name of the pool
            :param solution_ctg: the solution
            """
            for f in self.constraints_dict[constraints_name]:
                f(solution_ctg)

    def __init__(self, constraints_name):
        """
        It create the wrapper object which contains the chosen pool.
        Both global pools container object and required pool are created if they do not exist.

        :param constraints_name: The unique name of a pool.
        """
        if Constraints.__constrains is None:
            Constraints.__constrains = Constraints.__Constraints()
        _ = Constraints.__constrains.get_constraints(constraints_name)
        self.__constraints_name = constraints_name

    def __getattr__(self, name):
        return getattr(Constraints.__constrains, name)

    @property
    def constraints_name(self):
        """

        :return: The unique nome of the selected pool.
        """
        return self.__constraints_name

    def exec_constraints(self, solution_ctg):
        """
        It execute all the constraint functions on a given solution.

        :param solution_ctg: The solution in which execute the constraint functions.
        """
        self.__Constraints.exec_constraints(Constraints.__constrains, self.__constraints_name, solution_ctg)

    def insert_constraint(self, constraint_function):
        """
        It injects a new function inside the selected pool.
        All the functions must have only one input parameters (the solution),
        and they do not return anything, instead they raise the appropriate exception.

        :param constraint_function: The new function.
        """
        return self.__Constraints.insert_constraint(Constraints.__constrains, self.__constraints_name,
                                                    constraint_function)
