from abc import abstractmethod
from enum import Enum, unique
from random import randint

from microgp4 import microgp4
from microgp4.ctg import constrained_tagged_graph
from microgp4.ctg.constraints.constraints_type import FanInType
from microgp4.ctg.parameters.var_type import VarType
from microgp4.ctg.section import section_key
from microgp4.ctg.sectioned_tagged_graph import node_key
from microgp4.exception import Microgp4InvalidVariableType, Microgp4GenerationError


class ReferenceType(VarType):
    @abstractmethod
    def __str__(self):
        pass

    @staticmethod
    @abstractmethod
    def check_value(node_id, value):
        pass

    @staticmethod
    @abstractmethod
    def is_outer():
        pass

    @abstractmethod
    def get_label(self):
        pass

    @abstractmethod
    def mutate(self, strength):
        pass

    @staticmethod
    @abstractmethod
    def generate_new_value(node_id):
        pass


@unique
class _EdgeType(Enum):
    FORWARD = "inner_forward"
    BACKWARD = "inner_backward"
    SELF = "inner_self"
    OUT_SECTION = "outer"

    @staticmethod
    def __function_before(node, ctg):
        return ctg.get_previous_node(node)

    @staticmethod
    def __function_after(node, ctg):
        return ctg.get_successors_node(node)

    @staticmethod
    def __check_position_constraint(src, dst, get_node_function):

        ctg = microgp4.get_solution_by_node_id(src)
        node_to_check = get_node_function(src, ctg)
        section_node_list = ctg.section[ctg.node[src][section_key]][node_key]

        if dst not in ctg.node:
            return False

        if dst not in section_node_list:
            return False

        if dst not in node_to_check:
            return False

        return True

    @staticmethod
    def __check_before(src, dst):
        return _EdgeType.__check_position_constraint(src, dst, _EdgeType.__function_before)

    @staticmethod
    def __check_after(src, dst):
        return _EdgeType.__check_position_constraint(src, dst, _EdgeType.__function_after)

    def edge_valid(self, ctg, src, dst):

        name = str(self.name)

        if name == str(_EdgeType.FORWARD.name):
            return _EdgeType.__check_after(src, dst)
        if name == str(_EdgeType.BACKWARD.name):
            return _EdgeType.__check_before(src, dst)
        if name == str(_EdgeType.SELF.name):
            return src == dst
        if name == str(_EdgeType.OUT_SECTION.name):
            src_section = ctg.nodes[src][section_key]
            dst_section = ctg.nodes[dst][section_key]
            return src_section != dst_section
        return False

    def __str__(self):
        return self.value


def class_reference_type(edge_types=None, section=None, format_='{}', format_label='{}'):
    edge_valid_list = []

    if edge_types is None:
        edge_valid_list.append(_EdgeType.FORWARD)
        edge_valid_list.append(_EdgeType.BACKWARD)
        edge_valid_list.append(_EdgeType.SELF)
    else:
        for edge_type in edge_types:

            if isinstance(edge_type, _EdgeType):
                edge_enum = edge_type
            else:
                edge_enum = _EdgeType(edge_type)

                if edge_enum is _EdgeType.OUT_SECTION:
                    assert len(edge_valid_list) is 0
                    assert len(edge_types) is 1
                    assert section is not None

            edge_valid_list.append(edge_enum)

    is_outer_var = _EdgeType.OUT_SECTION in edge_valid_list

    class _ReferenceType(ReferenceType):

        __outer = "true" if is_outer_var else "false"

        def __init__(self, node_id, value=None):
            VarType.__init__(self, node_id, value)
            self._current_solution = microgp4.get_solution_by_node_id(node_id)
            self._node_section = self._current_solution.nodes[self.value[0]][section_key]

        @staticmethod
        def check_value(node_id, value):
            result = False
            current_solution = microgp4.get_solution_by_node_id(node_id)
            for r in edge_valid_list:
                result |= r.edge_valid(current_solution, node_id, value[0])
                if result:
                    break
            return result

        def __str__(self):
            return format_.format(self.value[0])

        @staticmethod
        def is_outer():
            return is_outer_var

        def get_label(self):
            return format_label.format(self.value[0])

        def mutate(self, strength):
            current_solution = microgp4.get_solution_by_node_id(self._node_id)

            if _EdgeType.OUT_SECTION not in edge_valid_list:
                nodes_candidate = list()

                if _EdgeType.BACKWARD in edge_valid_list:
                    nodes_candidate += current_solution.get_previous_node(self._node_id)
                if _EdgeType.SELF in edge_valid_list:
                    nodes_candidate.append(self._node_id)
                if _EdgeType.FORWARD in edge_valid_list:
                    nodes_candidate += current_solution.get_successors_node(self._node_id)

                val_position = nodes_candidate.index(self.value[0])

                new_position = randint(0, len(nodes_candidate))

                new_position -= val_position
                new_position *= strength
                new_position += val_position
                ref_node = nodes_candidate[new_position]

            else:
                section_list = current_solution.get_all_section_by_type(section)
                current_section = 0

                for s in section_list:
                    if self.value[0] in s[node_key]:
                        current_section = section_list.idex(s)
                        break

                section_position = randint(0, len(section_list))
                section_position -= current_section
                section_position *= strength
                section_position += current_section

                ref_node = section_list[section_position][constrained_tagged_graph.first_node_key]

            current_solution.remove_edge(self._node_id, self.value[0], self.value[1])

            key = current_solution.add_edge(self._node_id, ref_node, **{"type": "ref", "outer": _ReferenceType.__outer})
            self._value = (ref_node, key)

        @staticmethod
        def generate_new_value(node_id):

            current_solution = microgp4.get_solution_by_node_id(node_id)

            meta_section = current_solution.meta_section

            if _EdgeType.OUT_SECTION not in edge_valid_list:
                nodes_candidate = list()

                if _EdgeType.BACKWARD in edge_valid_list:
                    nodes_candidate += current_solution.get_previous_node(node_id)
                if _EdgeType.SELF in edge_valid_list:
                    nodes_candidate.append(node_id)
                if _EdgeType.FORWARD in edge_valid_list:
                    nodes_candidate += current_solution.get_successors_node(node_id)

                section_macros = \
                    meta_section.section[current_solution.section[current_solution.node[node_id]["section"]]["type"]][
                        "nodes"]["macros"]

                node_to_use = list()
                for node in nodes_candidate:
                    n = current_solution.nodes[node]
                    node_macro = n["macro"]
                    node_fan_in = section_macros[node_macro["group"]][node_macro["id"]]["fan_in"]
                    if node_fan_in is FanInType.INFINITE or current_solution.fan_in_node(node) < node_fan_in.value:
                        node_to_use.append(node)

                del nodes_candidate
                nodes_candidate = node_to_use
                del node_to_use

                if len(nodes_candidate) is 0:
                    raise Microgp4GenerationError("No available node.")

                ref_node = nodes_candidate[randint(0, len(nodes_candidate) - 1)]

            else:
                section_list = current_solution.get_all_section_by_type(section)
                section_fan_in = meta_section.section[section]["fan_in"]
                section_to_use = list()
                for section_id in section_list:
                    if section_fan_in is FanInType.INFINITE or current_solution.fan_in_section(
                            section_id) < section_fan_in.value:
                        section_to_use.append(section_id)

                del section_list
                section_list = section_to_use
                del section_to_use

                if len(section_list) is 0:
                    raise Microgp4GenerationError("No available section.")

                ref_node = section_list[randint(0, len(section_list) - 1)][constrained_tagged_graph.first_node_key]

            key = current_solution.add_edge(node_id, ref_node, **{"type": "ref", "outer": _ReferenceType.__outer})

            return ref_node, key

        @VarType.value.setter
        def value(self, value):

            if not isinstance(value, list):
                value = [value, ]

            if self.check_value(self._node_id, value):
                current_solution = microgp4.get_solution_by_node_id(self._node_id)
                if self.value is not None:
                    current_solution.remove_edge(self._node_id, self.value[0], self.value[1])

                key = current_solution.add_edge(self._node_id, value[0],
                                                **{"type": "ref", "outer": _ReferenceType.__outer})
                self._value = [value[0], key]
            else:
                self.log.error("The variable '%s' is not of the wanted type!" % value)
                raise Microgp4InvalidVariableType("Wrong variable type!")

    return _ReferenceType
