import logging
from abc import ABC, abstractmethod

from microgp4.exception import Microgp4InvalidVariableType


class VarType(ABC):
    """
    The abstract class which every parameters must extend.
    """
    def __init__(self, node_id, value=None):
        """
        If the value is given, it will check if it is a valid une.
        Otherwise a valid random one will be created

        :param node_id: The id of the node in which the parameters is contained
        :param value: the value of the parameters itself.
        """
        self.log = logging.getLogger("microgp4")
        self._node_id = node_id
        if value is None:
            self._value = self.generate_new_value(node_id)
        else:
            self._value = None
            self.value = value

    @abstractmethod
    def mutate(self, strength):
        """
        Mutate the internal value, bigger is the strength value and much more the new value must be
        different from the past.

        :param strength: float value between 0 and 1.
        :type strength: float
        """
        pass

    @staticmethod
    @abstractmethod
    def generate_new_value(node_id):
        """
        Generate a new object of the class within a random value.
        :param node_id: The id of the node in which the parameters is contained.
        """
        pass

    @staticmethod
    @abstractmethod
    def check_value(node_id, value):
        """
        Check if the value is a valid element

        :param node_id: The id of the node in which the parameters is contained.
        :param value: The value that has to be checked.
        :return: Boolean value representing the result of the check
        """
        pass

    def check(self):
        """
        Just a wrap function of the check_value.
        It will automatically pass the correct value to the function.

        :return: Boolean value representing the result of the check
        """
        return self.check_value(self._node_id, self.value)

    @property
    def value(self):
        """

        :return: return the value
        """
        return self._value

    @value.setter
    def value(self, value):
        """
        The given value is checked, before it substitutes the old one.

        :param value: The new value
        """
        if self.check_value(self._node_id, value):
            self._value = value
        else:
            self.log.error("The variable '%s' is not of the wanted type!" % value)
            raise Microgp4InvalidVariableType("Wrong variable type!")

    @abstractmethod
    def __str__(self):
        pass
