import logging

from microgp4.exception import Microgp4MethodError

section_key = "section"
sub_section_key = "sub_sections"
macros_key = "macros"


class Section:
    """
    Section class. It is one of the fundamental of the ConstrainedTaggedGraph class.
    This class is also used to create the object that contains the information read from the JSON files.
    """
    section_dict_factory = dict

    def __init__(self, data=None, **attr):
        self.log = logging.getLogger("microgp4")
        self.section_dict_factory = sdf = self.section_dict_factory
        self.section = sdf()
        if data is not None:
            self._add_attr_to_tags(data, **attr)
        else:
            data = attr
        self.section_data = data

    def _add_attr_to_tags(self, tags_dict, **attr):
        """
        Utility function to merge two tags together.

        :param tags_dict: The existing dict.
        :type tags_dict: dict
        :param attr: The new tag.
        """
        try:
            tags_dict.update(attr)
        except AttributeError:
            self.log.error("The tags_dict('%s') argument must be a dictionary." % tags_dict)
            raise Microgp4MethodError("The tags_dict argument must be a dictionary.")

    def add_section(self, section, tags_dict, **attr):
        """
        Method to create a new section, or update the content of an existing one.

        :param section: The section ID.
        :param tags_dict: The content of the section.
        :type tags_dict: dict
        :param attr: Additional values.
        """
        self._add_attr_to_tags(tags_dict, **attr)

        if section not in self.section:
            self.section[section] = tags_dict
        else:
            self.section[section].update(tags_dict)

    def remove_section(self, section, remove_all=False):
        """
        Remove section and optionally all the sub-section.

        :param section: The ID of the section that will be removed.
        :param remove_all: If True all the sub-sections are removed as well in recursive way.
        :type remove_all: bool
        """
        if section in self.section:

            if sub_section_key in self.section[section]:
                for s in self.section[section][sub_section_key]:
                    self.remove_section_from_section(section, s)
                    if remove_all:
                        self.remove_section(s)

            if section_key in self.section[section]:
                for s in self.section[section][section_key]:
                    self.remove_section_from_section(s, section)

            del self.section[section]

    def set_tags_to_section(self, section, tags_dict, **attr):
        """
        Set tags to section. The section must already exist.
        The new tag override completely the old information.

        :param section: The section ID.
        :param tags_dict: The new tag.
        :type tags_dict: dict
        :param attr: Additional information.
        """
        assert section in self.section
        assert tags_dict is not None

        self._add_attr_to_tags(tags_dict, **attr)
        self.section[section] = tags_dict

    def add_tags_to_section(self, section, tags_dict, **attr):
        """
        Set the tag if the Section is empty, otherwise it will add information.

        :param section: The Section ID.
        :param tags_dict: The new tag information.
        :type tags_dict: dict
        :param attr: Additional information.
        """
        assert section in self.section
        assert tags_dict is not None

        self._add_attr_to_tags(tags_dict, **attr)
        if self.section[section] is None:
            self.section[section] = tags_dict
        else:
            self.section[section].update(tags_dict)

    def remove_tags_from_section(self, section, tags_dict=None, **attr):
        """
        This method removes the entire tag or a particular field.
        If both tags_dict and attr are None, the entire tag will delete.
        If they contain keys only that values will be removed.

        :param section: The section ID.
        :param tags_dict: The tag containing the key to be removed or None to remove the entire tag.
        :type tags_dict: dict|None
        :param attr: Additional information.
        """
        assert section in self.section

        if tags_dict is None:
            tags_dict = attr
        else:
            self._add_attr_to_tags(tags_dict, **attr)

        if tags_dict is None:
            self.section[section] = None
        else:
            for k in tags_dict:
                if k in self.section[section]:
                    del self.section[section][k]

    def _check_if_section_is_inside_section(self, section, inner_section):
        """
        Utility method to check if section contains inner_section.
        If both or of the two sections do not exist the method will return False.

        :param section: The Section ID.
        :param inner_section: The ID of the Inner Section.
        :return: The boolean value representing if the inner_section if inside the section True, False otherwise.
        """
        if section not in self.section or inner_section not in self.section:
            return False
        outer_section = self.section[section]
        if sub_section_key in outer_section:
            for s in outer_section[sub_section_key]:
                if s["name"] is inner_section:
                    return True
        return False

    def get_sub_section(self, section, inner_section):
        """
        It will return the content of inner_section contained into section.

        :param section: The Section ID.
        :param inner_section: The ID of the Inner Section.
        :return: The content of inner_section.
        """

        assert section in self.section
        assert inner_section in self.section
        assert sub_section_key in self.section[section]

        for s in self.section[section][sub_section_key]:
            if s["name"] == inner_section:
                return s

        return None

    def put_section_inside_section(self, section, inner_section, **attr):
        """
        This method puts inner_section into section. Both sections must already exist.

        :param section: The Section ID.
        :param inner_section: The ID of the Inner Section.
        :param attr: Additional information for inner_section.
        """
        assert section, inner_section in self.section
        outer_section = self.section[section]

        if sub_section_key not in outer_section:
            outer_section[sub_section_key] = list()

        to_insert = {"name": inner_section}
        to_insert.update(attr)

        if inner_section not in outer_section[sub_section_key]:
            outer_section[sub_section_key].append(to_insert)
        else:
            self.log.warning("Section '%s' already has sub-section '%s' " % (section, inner_section))

        if section_key not in self.section[inner_section]:
            self.section[inner_section][section_key] = list()

        if section not in self.section[inner_section][section_key]:
            self.section[inner_section][section_key].append(section)
        else:
            self.log.warning("Section '%s' already has parent-section '%s' " % (inner_section, section))

    def remove_section_from_section(self, section, inner_section):
        """
        This method will remove inner_section from section, but no sections are deleted.
        This method removes only the information on which section is contained by what.
        Obviously the two sections must exist.

        :param section: The Section ID.
        :param inner_section: The ID of the Inner Section.
        """

        assert section in self.section and inner_section in self.section
        assert sub_section_key in self.section[section]

        sub_sec = self.section[section][sub_section_key]

        self.section[section][sub_section_key] = [s for s in sub_sec if s.get('name') != inner_section]
        self.section[inner_section][section_key].remove(section)
