from microgp4 import microgp4
from microgp4.ctg.section import Section, sub_section_key, section_key
from microgp4.ctg.tagged_graph import TaggedGraph

node_key = "node"


class SectionedTaggedGraph(TaggedGraph, Section):
    def __init__(self, data=None, **attr):
        TaggedGraph.__init__(self, data, **attr)
        Section.__init__(self)

    def add_section(self, tags_dict, **attr):
        """
        Add a new section into the graph, the ID will generate internally.
        It is not possible to override (or change) the ID.

        :param tags_dict: dictionary, optional (default= no attributes)
            Dictionary of node attributes.  Key/value pairs will
            update existing data associated with the node.
        :param attr: keyword arguments, optional
              Set or change attributes using key=value.
        :return: The Section ID.
        """
        section = microgp4.create_unique_section_code()
        assert section not in self.section
        self._add_attr_to_tags(tags_dict, **attr)
        self.section[section] = tags_dict

        return section

    def remove_section(self, section, remove_all=True):
        if section in self.section:

            if node_key in self.section[section]:
                for n in self.section[section][node_key]:
                    self.remove_node(n)
            self._remove_section(section, remove_all)

    def _remove_section(self, section, remove_all):
        if section in self.section:

            if sub_section_key in self.section[section]:
                for s in self.section[section][sub_section_key]:
                    self.remove_section_from_section(section, s)
                    if remove_all:
                        self.remove_section(s)

            if section_key in self.section[section]:
                self.remove_section_from_section(self.section[section][section_key], section)

            del self.section[section]

    def put_section_inside_section(self, section, inner_section):
        assert section in self.section and inner_section in self.section
        outer_section = self.section[section]

        if sub_section_key not in outer_section:
            outer_section[sub_section_key] = list()

        if section not in outer_section[sub_section_key]:
            outer_section[sub_section_key].append(inner_section)
        else:
            self.log.warning("Section '%s' already has sub-section '%s' " % (section, inner_section))

        inner_attr = {section_key: section}

        self.add_tags_to_section(inner_section, inner_attr)

    def remove_section_from_section(self, section, inner_section):
        assert section in self.section and inner_section in self.section
        assert sub_section_key in self.section[section]

        self.section[section][sub_section_key].remove(inner_section)
        self.remove_tags_from_section(inner_section, {section_key})

    def add_node(self, section, **attr):
        """Adding new node, put it inside a section and update node attributes.

           Parameters
          ----------

          section : section
                      The section id in which the node will be inserted.
          attr : keyword arguments, optional
              Set or change attributes using key=value.
          :return the node name"""

        assert section in self.section
        node = TaggedGraph.add_node(self, **attr)
        self.put_node_inside_section(section, node)
        return node

    def remove_node(self, node):
        if node in self.node:
            if section_key in self.nodes[node]:
                section = self.nodes[node][section_key]
                self.remove_node_from_section(section, node)
        TaggedGraph.remove_node(self, node)

    def put_node_inside_section(self, section, node):
        assert section in self.section
        assert node in self.node

        if node_key not in self.section[section]:
            self.section[section][node_key] = list()

        if node not in self.section[section][node_key]:
            self.section[section][node_key].append(node)
        else:
            self.log.warning("Section '%s' already has node '%s' " % (section, node))

        self.add_tags_to_node(node, {section_key: section})

    def remove_node_from_section(self, section, node):
        assert section in self.section
        assert node in self.node

        self.section[section][node_key].remove(node)
        self.remove_tags_fom_node(node, {section_key})
