import logging

from networkx import MultiDiGraph

from microgp4 import microgp4
from microgp4.exception import Microgp4MethodError


class TaggedGraph(MultiDiGraph):
    """Add better support for attr (become tags) extended all the functionality of the MultiDiGraph"""

    def __init__(self, data=None, **attr):
        self.log = logging.getLogger("microgp4")
        MultiDiGraph.__init__(self, data, **attr)

    def _create_unique_node_code(self):
        code = microgp4.create_unique_node_code()
        assert code not in self.nodes
        return code

    def _create_unique_edge_code(self, u, v):
        code = microgp4.create_unique_edge_code()
        try:
            edg = self.adj[u][v]
        except KeyError:
            return code
        assert code not in edg
        return code

    def set_tags_to_graph(self, tags_dict, **attr):
        assert tags_dict is not None
        self._add_attr_to_tags(tags_dict, **attr)
        self.graph = tags_dict

    def add_tags_to_graph(self, tags_dict, **attr):
        assert tags_dict is not None

        self._add_attr_to_tags(tags_dict, **attr)

        if self.graph is not None:
            self.graph.update(tags_dict)
        else:
            self.graph = tags_dict

    def remove_tags_from_graph(self, tags_dict, **attr):
        if tags_dict is None:
            tags_dict = attr
        else:
            self._add_attr_to_tags(tags_dict, **attr)

        if tags_dict is None:
            self.graph = None
        else:
            for k in tags_dict:
                if k in self.graph:
                    del self.graph[k]

    def add_node(self, **attr):
        """Add a single node and update node attributes.

         Parameters
        ----------
        attr : keyword arguments, optional
            Set or change attributes using key=value.

        :return the node name"""

        n = self._create_unique_node_code()
        MultiDiGraph.add_node(self, n, **attr)
        return n

    def add_nodes_from(self, nodes, **attr):
        assert nodes not in self.nodes
        MultiDiGraph.add_nodes_from(nodes, **attr)

    def _add_attr_to_tags(self, tags_dict, **attr):
        try:
            tags_dict.update(attr)
        except AttributeError:
            self.log.error("The tags_dict('%s') argument must be a dictionary." % tags_dict)
            raise Microgp4MethodError("The tags_dict argument must be a dictionary.")

    def set_tags_to_node(self, node, tags_dict, **attr):
        assert node in self.nodes
        assert tags_dict is not None

        self._add_attr_to_tags(tags_dict, **attr)

        self._node[node] = tags_dict

    def add_tags_to_node(self, node, tags_dict, **attr):
        assert node in self.nodes
        assert tags_dict is not None

        self._add_attr_to_tags(tags_dict, **attr)

        if self.nodes[node] is not None:
            self.nodes[node].update(tags_dict)
        else:
            self._node[node] = tags_dict

    def remove_tags_fom_node(self, node, tags_dict=None, **attr):
        assert node in self.nodes

        if tags_dict is None:
            tags_dict = attr
        else:
            self._add_attr_to_tags(tags_dict, **attr)

        if tags_dict is None:
            self._node[node] = None
        else:
            for k in tags_dict:
                if k in self.nodes[node]:
                    del self.nodes[node][k]

    def add_edge(self, u, v, **attr):
        """Add an edge between u and v.

        The nodes u and v must already exist in the graph.

        Edge attributes can be specified with keywords or by providing
        a dictionary with key/value pairs.

        Parameters
        ----------
        u, v : nodes
        attr : keyword arguments, optional
            Edge data (or labels or objects) can be assigned using
            keyword arguments.

        :return the edge name"""

        assert u in self.nodes and v in self.nodes
        key = self._create_unique_edge_code(u, v)
        MultiDiGraph.add_edge(self, u, v, key, **attr)
        return key

    def _check_edge_exist(self, u, v, edge_name):
        assert u, v in self.nodes
        assert u in self.adj
        assert v in self.adj[u]
        assert edge_name in self.adj[u][v]

    def set_tags_to_edge(self, u, v, edge_name, tags_dict, **attr):
        self._check_edge_exist(u, v, edge_name)
        assert tags_dict is not None

        self._add_attr_to_tags(tags_dict, **attr)
        self.adj[u][v][edge_name] = tags_dict

    def add_tags_to_edge(self, u, v, edge_name, tags_dict, **attr):
        self._check_edge_exist(u, v, edge_name)
        assert tags_dict is not None

        self._add_attr_to_tags(tags_dict, **attr)

        if self.adj[u][v][edge_name] is None:
            self.adj[u][v][edge_name] = tags_dict
        else:
            self.adj[u][v][edge_name].update(tags_dict)

    def remove_tags_from_edge(self, u, v, edge_name, tags_dict, **attr):
        self._check_edge_exist(u, v, edge_name)
        assert tags_dict is not None

        if tags_dict is None:
            tags_dict = attr
        else:
            self._add_attr_to_tags(tags_dict, **attr)

        if tags_dict is None:
            self.adj[u][v][edge_name] = None
        else:
            for k in tags_dict:
                if k in self.adj[u][v][edge_name]:
                    del self.adj[u][v][edge_name][k]
