import json
import logging
import os
from itertools import groupby

from microgp4.ctg.constraints.constraints_type import FanInType
from microgp4.ctg.parameters import reference_type
from microgp4.ctg.parameters.bit_vector_type import class_bit_vector
from microgp4.ctg.parameters.enum_type import class_enum
from microgp4.ctg.parameters.float_type import class_float_range
from microgp4.ctg.parameters.integer_type import class_inter_range
from microgp4.ctg.parameters.special_type import ClassSelf
from microgp4.ctg.section import Section, macros_key, sub_section_key
from microgp4.exception import Microgp4UndefinedType, Microgp4UndefinedSection, Microgp4InvalidVariable, \
    Microgp4InvalidSection, Microgp4InvalidMacro


class JsonReader:
    def __init__(self, file_path):
        self.__log = logging.getLogger("microgp4")
        self.__types = {"self": ClassSelf}
        self.__macros = dict()
        self.__sections = dict()
        self.__json_path = file_path
        self.__json_folder = os.path.dirname(file_path)
        self.__raw_data = self._read_file()
        self.__ref_sections = list()

    def _read_file(self, json_file_path=None):
        """Simply read data from JSON file without any elaboration.


        Parameters
        ----------
        json_file_path: str
            Absolute (or relative for the program) path contained the JSON file.

        Returns
        -------
        dict
            The JSON structure as dictionary.
        """
        if json_file_path is None:
            file_path = self.__json_path
        else:
            file_path = json_file_path
        with open(file_path) as data_file:
            return json.load(data_file)

    def _create_type(self, parameter):
        type_ = parameter["type"]
        format_ = '{}'

        if "format" in parameter:
            format_ = parameter["format"]
        if type_ == 'integer':
            return class_inter_range(int(parameter["min"]), int(parameter["max"]), format_)
        elif type_ == 'float':
            return class_float_range(float(parameter["min"]), float(parameter["max"]), format_)
        elif type_ == 'enum':
            return class_enum(tuple(parameter["elements"]), format_)
        elif type_ == 'bit_vector':
            return class_bit_vector(parameter["length"], format_)
        elif type_ == 'inner_reference':
            edge_types = []

            section = None
            if 'valid_destinations' not in parameter:
                edge_types = None
            else:
                destinations = parameter['valid_destinations']
                if isinstance(destinations, str):
                    edge_types.append(destinations)
                elif isinstance(destinations, list):
                    edge_types = destinations

                for i, e in enumerate(edge_types):
                    edge_types[i] = "inner_" + e

            format_label = "{}"

            if "format_label" in parameter:
                format_label = parameter["format_label"]

            return reference_type.class_reference_type(edge_types=edge_types, section=section, format_=format_,
                                                       format_label=format_label)
        elif type_ == 'outer_reference':
            edge_types = ["outer"]
            section = parameter['section']
            self.__ref_sections.append(section)
            return reference_type.class_reference_type(edge_types=edge_types, section=section, format_=format_)
        else:
            self.__log.fatal("Type '%s' does not exist!", type_)
            raise Microgp4UndefinedType("Type '%s' does not exist!" % type_)

    def _create_macro(self, macro):
        parameters = dict()

        if "parameters" in macro:
            for name_v, p in macro["parameters"].items():
                if type(p) is dict:
                    parameters[str(name_v)] = self._create_type(p)
                else:
                    parameters[str(name_v)] = str(p)

        molt = 1
        fan_in = FanInType.INFINITE

        if 'molt' in macro:
            molt = int(macro['molt'])

        if "fan_in" in macro:
            try:
                fan_in = FanInType(macro['fan_in'])
            except ValueError:
                error_text = "Parameter '{}' is not valid! Expect 0, 1 or \"inf\". ".format(macro['fan_in'])
                self.__log.fatal(error_text)
                raise Microgp4InvalidVariable(error_text)

        result = {"parameters": parameters, "expression": macro["expression"], "molt": molt, "fan_in": fan_in}

        if "name" in macro:
            result["name"] = macro["name"]

        return result

    @staticmethod
    def __find_between(s, first, last):
        result = []
        s_len = len(s)
        end = 0
        try:
            while s_len >= end:
                start = s.index(first, end) + len(first)
                end = s.index(last, start)
                result.append(s[start:end])
            return result
        except ValueError:
            return result

    @staticmethod
    def _find_macro_reference(s, symbol='#'):
        return JsonReader.__find_between(s, '%s{' % symbol, '}')

    @staticmethod
    def _find_macro_parameters(s, symbol='$'):
        return JsonReader.__find_between(s, '%s{' % symbol, '}')

    @staticmethod
    def __section_min_max(section_dic):
        _min = "min"
        _max = "max"
        min_max = dict()

        min_max[_min] = 1
        min_max[_max] = 42000000000
        if _min in section_dic:
            min_max[_min] = section_dic[_min]
        if _max in section_dic:
            min_max[_max] = section_dic[_max]

        assert min_max[_min] <= min_max[_max]
        assert min_max[_max] > 0
        assert min_max[_min] >= 0

        return min_max

    def _create_section(self, section):
        _min = "min"
        _max = "max"
        _sub_sections = "sub_sections"
        _nodes = "nodes"
        _name = "name"
        _macros = "macros"
        section_result = dict()

        assert _sub_sections not in section or _nodes not in self.__sections
        assert _nodes not in self.__sections or _sub_sections not in self.__sections

        min_max = self.__section_min_max(section)

        if "fan_in" in section:
            try:
                section_result['fan_in'] = FanInType(section['fan_in'])
            except ValueError:
                error_text = "Parameter '{}' is not valid! Expect 0, 1 or \"inf\". ".format(section['fan_in'])
                self.__log.fatal(error_text)
                raise Microgp4InvalidVariable(error_text)
        else:
            section_result['fan_in'] = FanInType.INFINITE

        section_result[_max] = min_max[_max]
        section_result[_min] = min_max[_min]

        if _sub_sections in section:
            section_result[_sub_sections] = list()
            sub_list = section_result[_sub_sections]
            sub_s = section[_sub_sections]
            if type(sub_s) is str:
                sub_list.append({_name: sub_s, _min: 1, _max: 1})
            elif type(sub_s) is list:
                for sb in sub_s:
                    if type(sb) is str:
                        sub_list.append({_name: sb, _min: 1, _max: 1})
                    elif type(sb) is dict:
                        sb_section = self.__section_min_max(sb)
                        sb_section[_name] = sb[_name]
                        sub_list.append(sb_section)
        elif _nodes in section:
            nodes = section[_nodes]
            node = self.__section_min_max(nodes)
            assert node[_min] > 0
            mcs = nodes[_macros]
            if type(mcs) is str:
                node[_macros] = list()
                node[_macros].append(mcs)
            elif type(mcs) is list:
                node[_macros] = mcs
            section_result[_nodes] = node

        return section_result

    def _parsing_data(self, data):
        if "include_files" in data:

            include_files = data["include_files"]
            for in_f in include_files:
                if os.path.isabs(in_f):
                    sub_file_path = in_f
                else:
                    sub_file_path = os.path.normpath(os.path.join(self.__json_folder, in_f))
                self._parsing_data(self._read_file(sub_file_path))

        if 'types' in data:
            tps = data['types']
            for name, p in tps.items():
                name_str = str(name)

                if name_str in self.__types:
                    error_message = "Invalid type! '{}' is already present!".format(name_str)
                    self.__log.fatal(error_message)
                    raise Microgp4InvalidVariable(error_message)

                self.__types[name_str] = self._create_type(p)

        if "macros" in data:
            mcs = data["macros"]
            for group, macros_list in mcs.items():
                default_key = 0
                if group in self.__macros:
                    error_message = "Macro group '{}' already present!".format(group)
                    self.__log.fatal(error_message)
                    raise Microgp4InvalidMacro(error_message)
                self.__macros[group] = dict()
                mcs_group = self.__macros[group]
                if type(macros_list) is str:
                    mcs_group[default_key] = {"parameters": dict(), "expression": macros_list, 'molt': 1,
                                              'group': group, "fan_in": FanInType.INFINITE}
                    default_key += 1

                elif type(macros_list) is dict:
                    result = self._create_macro(macros_list)
                    result['group'] = group
                    mcs_group[result.pop("name", default_key)] = result
                    default_key += 1

                elif type(macros_list) is list:
                    for m in macros_list:
                        if type(m) is str:
                            mcs_group[default_key] = {"parameters": dict(), "expression": m, 'molt': 1,
                                                      'group': group, "fan_in": FanInType.INFINITE}
                            default_key += 1

                        elif type(m) is dict:
                            result = self._create_macro(m)
                            result['group'] = group
                            mcs_group[result.pop("name", default_key)] = result

                            default_key += 1

        if "sections" in data:
            scs = data["sections"]

            for section_name, section in scs.items():
                if section_name in self.__sections:
                    error_message = "Section {} is duplicated!".format(section_name)
                    self.__log.fatal(error_message)
                    raise Microgp4InvalidSection(error_message)
                self.__sections[section_name] = self._create_section(section)
                if section_name == "root":
                    self.__sections["root"]['fan_in'] = FanInType.ZERO

    def _build_macro(self, macro_dict):
        for _, macro in macro_dict.items():
            exp = macro["expression"]

            parameters = macro["parameters"]

            par_symbol = macro.pop("par_symbol", '$')
            ref_symbol = macro.pop("ref_symbol", '#')

            exp_par = self._find_macro_parameters(exp, par_symbol)
            used_parameters = dict()

            exp_ref = self._find_macro_reference(exp, ref_symbol)

            if len(exp_par) > 0:
                for i, p in enumerate(exp_par):
                    exp = exp.replace(par_symbol + '{' + p + '}', '{' + par_symbol + p + '}')
                    if type(parameters[p]) is str:
                        parameters[p] = self.__types[parameters[p]]
                    par_class = parameters.pop(p)
                    exp_par[i] = par_symbol + p
                    used_parameters[exp_par[i]] = par_class
                macro["expression"] = exp

            macro["parameters"] = used_parameters

            if len(exp_ref) > 0:
                exp_ref.sort()
                pars_occurrence = dict()

                for key, group in groupby(exp_ref):
                    pars_occurrence[key] = len(list(group))

                for key, value in pars_occurrence.items():
                    if value > 1:
                        for i in range(0, value):
                            unique_par = key + str(i)
                            position = exp_ref.index(key)
                            exp_ref[position] = unique_par
                            exp = exp.replace(ref_symbol + '{' + key + '}', '{' + unique_par + '}', 1)
                            macro["parameters"][unique_par] = self.__types[key]
                    elif value is 1:
                        exp = exp.replace(ref_symbol + '{' + key + '}', '{' + key + '}')
                        macro["parameters"][key] = self.__types[key]
                    macro["expression"] = exp

    def _build_sections(self):
        section = Section()
        to_put_in_section = dict()
        for s_name, s in self.__sections.items():
            tag = {"min": s["min"], "max": s["max"], "fan_in": s["fan_in"]}
            if "nodes" in s:
                tag["nodes"] = dict()
                tag["nodes"]["min"] = s["nodes"]["min"]
                tag["nodes"]["max"] = s["nodes"]["max"]
                tag["nodes"][macros_key] = dict()
                for m in s["nodes"][macros_key]:
                    tag["nodes"][macros_key][m] = self.__macros[m]
            else:
                to_put_in_section[s_name] = s[sub_section_key]

            section.add_section(s_name, tag)

        for s, sub_s_list in to_put_in_section.items():
            for sub_s in sub_s_list:
                tag = {"min": sub_s["min"], "max": sub_s["max"]}
                section.put_section_inside_section(s, sub_s["name"], **tag)

        return section

    def build_constraints(self):
        self._parsing_data(self.__raw_data)
        assert len(self.__macros) > 0
        assert len(self.__sections) > 0

        if len(self.__ref_sections) > 0:
            for s in self.__ref_sections:
                if s not in self.__sections:
                    error_text = "Referenced section '{}' does not exist!".format(s)
                    self.__log.fatal(error_text)
                    raise Microgp4UndefinedSection(error_text)
                if "nodes" not in self.__sections[s]:
                    error_text = "Referenced section '{}' can not have nodes!".format(s)
                    self.__log.fatal(error_text)
                    raise Microgp4InvalidSection(error_text)

        for _, m in self.__macros.items():
            self._build_macro(m)

        return self.__macros, self._build_sections()
