import logging
from random import randint, choice

from microgp4.ctg.constrained_tagged_graph import ConstrainedTaggedGraph
from microgp4.ctg.constraints.constraints import Constraints
from microgp4.exception import Microgp4Exception
from microgp4.json import json

__log = logging.getLogger("microgp4")

__unique_codes = {}
__solutions = []
__node_pools = dict()


def __create_unique_code(code_type=""):
    if any((code_type == "", code_type is None)):
        code_type = "default_node_type"

    if code_type not in __unique_codes:
        __unique_codes[code_type] = 0

    ret = __unique_codes[code_type]
    __unique_codes[code_type] += 1
    return ret


def create_unique_node_code():
    return "n_{}".format(__create_unique_code("node"))


def create_unique_edge_code():
    return "e_{}".format(__create_unique_code("edge"))


def create_unique_section_code():
    return "s_{}".format(__create_unique_code("section"))


def get_solution_by_node_id(node_id):
    if len(__solutions) > 0:
        for s in __solutions:
            nodes = s.node
            if node_id in nodes:
                return s

    error_message = "Impossible to find solution containing node id: '{0}'!".format(node_id)
    __log.exception(error_message)
    raise Microgp4Exception(error_message)


def add_new_solution(solution):
    assert isinstance(solution, ConstrainedTaggedGraph)
    __solutions.append(solution)


def remove_solution(solution):
    __solutions.remove(solution)


def __add_sub_section(solution, section_parent_id, section_c, all_section):
    section_attr = all_section[section_c['name']]
    section_type = section_c['name']
    n_new_section = randint(section_c['min'], section_c['max'])
    for i in range(0, n_new_section):
        section_id = solution.add_section(section_type)
        solution.put_section_inside_section(section_parent_id, section_id)

        if 'sub_sections' in section_attr:
            for sub_section in section_attr['sub_sections']:
                __add_sub_section(solution, section_id, sub_section, all_section)
        elif 'nodes' in section_attr:
            nodes = section_attr['nodes']
            nodes_number = randint(nodes['min'], nodes['max'])

            if section_type not in __node_pools:
                macro_pool = []

                for k, e in nodes['macros'].items():
                    for node_id, m in e.items():
                        molt = m['molt']
                        for _ in range(0, molt):
                            macro_pool.append({'macro': m, 'node_id': node_id})

                __node_pools[section_type] = macro_pool

            else:
                macro_pool = __node_pools[section_type]

            for j in range(0, nodes_number):
                rnd_macro = choice(macro_pool)
                solution.append_node(section_id, rnd_macro['macro']['group'], rnd_macro['node_id'])
        else:
            __log.warning("Found empty section of type: '{0}'! It will be removed!".format(section_c['name']))
            solution.remove_section(section_id)


def create_solution(json_path, constraints_name='default_name', solution_number=1):
    assert solution_number > 0 and isinstance(solution_number, int)
    j_reader = json.JsonReader(json_path)
    c = Constraints(constraints_name)
    # here it is possible to add constraints to c
    _, s = j_reader.build_constraints()
    root_section_c = s.section['root']
    solutions = list()
    for i in range(0, solution_number):
        new_ctg = ConstrainedTaggedGraph(c.constraints_name, s)
        __solutions.append(new_ctg)

        try:
            r_id = new_ctg.add_section('root')
            for sub_section in root_section_c['sub_sections']:
                __add_sub_section(new_ctg, r_id, sub_section, s.section)
            new_ctg.finalize_nodes()
        except Microgp4Exception as err:
            __log.exception("{0}".format(err))
            __solutions.remove(new_ctg)
            return None
        if not new_ctg.check():
            __solutions.remove(new_ctg)
            return None

        solutions.append(new_ctg)

    return solutions
